// Akos.cpp
// By Nolan Check

#include "Akos.h"
#include "Scumm.h"
#include "BitStream.h"

bool Akos::Initialize()
{
	BLOCKHEADER bh;
	bh = *(BLOCKHEADER*)m_pAkosData;
	if (bh.dwName != BEDW('AKOS'))
	{
		ErrorMessage("Invalid AKOS block header.");
		return false;
	}

	DWORD dwAkosSize = BEDW(bh.dwSize);
	DWORD dwCursor = 8;
	do
	{
		bh = *(BLOCKHEADER*)(m_pAkosData + dwCursor);
		dwCursor += 8;

		switch (bh.dwName)
		{
		case BEDW('AKHD'):
			m_pAKHD = (AKHDBLOCK*)(m_pAkosData + dwCursor);
			break;
		case BEDW('AKPL'):
			m_dwNumColors = BEDW(bh.dwSize) - 8;
			m_pAKPL = m_pAkosData + dwCursor;
			break;
		case BEDW('RGBS'):
			m_pRGBS = m_pAkosData + dwCursor;
			break;
		case BEDW('AKSQ'):
			m_pAKSQ = m_pAkosData + dwCursor;
			break;
		case BEDW('AKCH'):
			m_pAKCH = m_pAkosData + dwCursor;
			break;
		case BEDW('AKOF'):
			m_pAKOF = (AKOFENTRY*)(m_pAkosData + dwCursor);
			break;
		case BEDW('AKCI'):
			m_pAKCI = m_pAkosData + dwCursor;
			break;
		case BEDW('AKCD'):
			m_pAKCD = m_pAkosData + dwCursor;
			break;
		default:
			ErrorMessage("Unknown block found in AKOS.");
			return false;
		}

		dwCursor += BEDW(bh.dwSize) - 8;
	} while (dwCursor < dwAkosSize);

	return true;
}

bool Akos::LoadFrame(WORD wFrame)
{
	if (!wFrame)
	{
		ErrorMessage("Attempted to load frame 0.");
		return false;
	}
	if (wFrame > m_pAKHD->wNumFrames)
	{
		ErrorMessage("Attempted to load non-existent frame.");
		return false;
	}

	m_pCurrentAKCD = m_pAKCD + m_pAKOF[wFrame - 1].dwAKCD;
	m_pCurrentAKCI = (AKCIENTRY*)(m_pAKCI + m_pAKOF[wFrame - 1].wAKCI);
	m_wCurrentFrame = wFrame;

	return true;
}

bool Akos::DrawFrame(Sprite& sprite, int x, int y)
{
	if (!m_wCurrentFrame)
	{
		return false;
	}

	if (!sprite.Initialize(x, y, m_pCurrentAKCI->wWidth, m_pCurrentAKCI->wHeight, 256))
	{
		return false;
	}

	memcpy(sprite.GetPalette(), m_pRGBS, m_dwNumColors * 3);

	switch (m_pAKHD->wCodec)
	{
	case 1:
		Codec1(sprite.GetBits());
		break;
	case 5:
		Codec5(sprite.GetBits());
		break;
	case 16:
		Codec16(sprite.GetBits());
		break;
	default:
		ErrorMessage("Unknown codec.");
		sprite.Destroy();
		return false;
	}

	return true;
}

void Akos::Codec1(BYTE* pDst)
{
	int  iShift;
	BYTE bMask;
	switch (m_dwNumColors)
	{
	case 32:
		iShift = 3;
		bMask  = 0x7;
		break;
	case 64:
		iShift = 2;
		bMask  = 0x3;
		break;
	default:
		iShift = 4;
		bMask  = 0xF;
		break;
	}

	int   curX      = 0;
	WORD  wWidth    = m_pCurrentAKCI->wWidth;
	WORD  wHeight   = m_pCurrentAKCI->wHeight;
	BYTE* pDestData = pDst;
	BYTE* pSrcData  = m_pCurrentAKCD;

	while (1)
	{
		BYTE bCode = *pSrcData++;
		BYTE bColor = bCode >> iShift;
		BYTE bRep = bCode & bMask;

		if (!bRep)
		{
			bRep = *pSrcData++;
		}

		for (BYTE b = 0; b < bRep; b++)
		{
			*pDestData = bColor;
			pDestData += m_pCurrentAKCI->wWidth;

			if (!--wHeight)
			{
				wHeight = m_pCurrentAKCI->wHeight;
				curX++;
				pDestData = pDst + curX;

				if (!--wWidth)
				{
					return;
				}
			}
		}
	}
}

void Akos::Codec5(BYTE* pDst)
{
	BYTE* pSrc    = m_pCurrentAKCD;
	WORD  wHeight = m_pCurrentAKCI->wHeight;

	do
	{
		WORD wLineSize = *(WORD*)pSrc;
		pSrc += 2;

		BYTE* pCurSrc = pSrc;
		BYTE* pCurDst = pDst;

		WORD wLen = m_pCurrentAKCI->wWidth;
		while (wLen > 0)
		{
			BYTE bCode = *pCurSrc++;

			BYTE bNum = (bCode >> 1) + 1;
			if (bNum > wLen)
			{
				bNum = (BYTE)wLen;
			}
			wLen -= bNum;

			if (bCode & 1)
			{
				BYTE bColor = *pCurSrc++;
				memset(pCurDst, bColor, bNum);
			}
			else
			{
				for (BYTE b = 0; b < bNum; b++)
				{
					BYTE bColor = *pCurSrc++;
					pCurDst[b] = bColor;
				}
			}

			pCurDst += bNum;
		}

		pSrc += wLineSize;
		pDst += m_pCurrentAKCI->wWidth;
	} while (--wHeight);
}

void Akos::Codec16(BYTE* pDst)
{
	BYTE bColorBits = *m_pCurrentAKCD;
	BYTE bColor = *(m_pCurrentAKCD + 1);

	BitStream stream(m_pCurrentAKCD + 2);

	WORD wHeight = m_pCurrentAKCI->wHeight;
	BYTE bRep = 0;
	while (wHeight--)
	{
		WORD wWidth = m_pCurrentAKCI->wWidth;
		while (wWidth--)
		{
			*pDst++ = bColor;

			if (bRep)
			{
				bRep--;
				continue;
			}

			if (stream.ReadBits(1))
			{
				if (stream.ReadBits(1))
				{
					BYTE bCode = stream.ReadBits(3);
					if (bCode == 4)
					{
						bRep = stream.ReadBits(8) - 1;
					}
					else
					{
						bColor = bCode - 4;
					}
				}
				else
				{
					bColor = stream.ReadBits(bColorBits);
				}
			}
		}
	}
}