// Display.cpp
// By Nolan Check

#include "Display.h"
#include <math.h>

#define DISPLAY_WIDTH  640
#define DISPLAY_HEIGHT 480
#define DISPLAY_DEPTH  8
#define DISPLAY_COLORS 256

bool Display::Initialize(HWND hWindow)
{
	if (m_fReady)
	{
		Destroy();
	}

	m_hWindow = hWindow;

	HDC hdc = GetDC(m_hWindow);
	m_hdcDisplay = CreateCompatibleDC(hdc);
	ReleaseDC(m_hWindow, hdc);

	BITMAPINFO* pbi;
	pbi = (BITMAPINFO*)malloc(sizeof(BITMAPINFOHEADER) + (sizeof(RGBQUAD) * DISPLAY_COLORS));
	if (!pbi)
	{
		ErrorMessage("Out of memory.");
		return false;
	}
	pbi->bmiHeader.biSize          = sizeof(BITMAPINFOHEADER);
	pbi->bmiHeader.biWidth         = DISPLAY_WIDTH;
	pbi->bmiHeader.biHeight        = -DISPLAY_HEIGHT;
	pbi->bmiHeader.biPlanes        = 1;
	pbi->bmiHeader.biBitCount      = DISPLAY_DEPTH;
	pbi->bmiHeader.biCompression   = BI_RGB;
	pbi->bmiHeader.biSizeImage     = DISPLAY_WIDTH * DISPLAY_HEIGHT;
	pbi->bmiHeader.biXPelsPerMeter = 0;
	pbi->bmiHeader.biYPelsPerMeter = 0;
	pbi->bmiHeader.biClrUsed       = 0;
	pbi->bmiHeader.biClrImportant  = 0;

	ZeroMemory(pbi->bmiColors, sizeof(RGBQUAD) * DISPLAY_COLORS);
	pbi->bmiColors[255].rgbRed   = 255;
	pbi->bmiColors[255].rgbGreen = 255;
	pbi->bmiColors[255].rgbBlue  = 255;

	m_hbmDisplay = CreateDIBSection(m_hdcDisplay, pbi, DIB_RGB_COLORS, (void**)&m_pBits, NULL, 0);
	free(pbi);
	if (!m_hbmDisplay)
	{
		ErrorMessage("Failed to create DIBSection.");
		return false;
	}

	m_hbmOld = (HBITMAP)SelectObject(m_hdcDisplay, m_hbmDisplay);

	m_fReady = true;
	return true;
}

void Display::Destroy()
{
	if (m_fReady)
	{
		SelectObject(m_hdcDisplay, m_hbmOld);
		DeleteObject(m_hbmDisplay);
		DeleteDC(m_hdcDisplay);
		m_fReady = false;
	}
}

void Display::SetPalette(BYTE* pPalette, UINT numColors)
{
	RGBQUAD rgbq[DISPLAY_COLORS];
	for (UINT i = 0; i < numColors; i++)
	{
		rgbq[i].rgbRed      = *pPalette++;
		rgbq[i].rgbGreen    = *pPalette++;
		rgbq[i].rgbBlue     = *pPalette++;
		rgbq[i].rgbReserved = 0;
	}
	rgbq[255].rgbRed   = 255;
	rgbq[255].rgbGreen = 255;
	rgbq[255].rgbBlue  = 255;

	SetDIBColorTable(m_hdcDisplay, 0, numColors, rgbq);
}

void Display::Clear()
{
	RECT rc;
	GetClientRect(m_hWindow, &rc);
	FillRect(m_hdcDisplay, &rc, (HBRUSH)GetStockObject(BLACK_BRUSH));
	InvalidateRect(m_hWindow, &rc, FALSE);
}

void Display::DrawSprite(Sprite& sprite)
{
	SetPalette(sprite.GetPalette(), sprite.GetNumColors());

	if (sprite.GetX() + sprite.GetWidth() < 0)
	{
		return;
	}
	if (sprite.GetX() > DISPLAY_WIDTH)
	{
		return;
	}
	if (sprite.GetY() + sprite.GetHeight() < 0)
	{
		return;
	}
	if (sprite.GetY() > DISPLAY_HEIGHT)
	{
		return;
	}

	WORD wDrawWidth = sprite.GetWidth();
	WORD wDrawHeight = sprite.GetHeight();

	int drawX = 0;
	if (sprite.GetX() < 0)
	{
		drawX = abs(sprite.GetX());
		wDrawWidth -= drawX;
	}
	if (sprite.GetX() + sprite.GetWidth() > DISPLAY_WIDTH)
	{
		wDrawWidth -= (sprite.GetX() + sprite.GetWidth()) - DISPLAY_WIDTH;
	}
	int drawY = 0;
	if (sprite.GetY() < 0)
	{
		drawY = abs(sprite.GetY());
		wDrawHeight -= drawY;
	}
	if (sprite.GetY() + sprite.GetHeight() > DISPLAY_HEIGHT)
	{
		wDrawHeight -= (sprite.GetY() + sprite.GetHeight()) - DISPLAY_HEIGHT;
	}

	BYTE* pSrc = sprite.GetBits() + drawX + (drawY * sprite.GetWidth());
	BYTE* pDst = m_pBits + (sprite.GetX() < 0 ? 0 : sprite.GetX()) + ((sprite.GetY() < 0 ? 0 : sprite.GetY()) * DISPLAY_HEIGHT);
	for (WORD w = 0; w < wDrawHeight; w++)
	{
		memcpy(pDst, pSrc, wDrawWidth);
		pDst += DISPLAY_WIDTH;
		pSrc += sprite.GetWidth();
	}

	RECT rc;
	GetClientRect(m_hWindow, &rc);
	InvalidateRect(m_hWindow, &rc, FALSE);
}