// Scumm.cpp
// By Nolan Check

#include "Scumm.h"

bool ScummDisk::Initialize(File* pFile)
{
	BLOCKHEADER bh;
	if (!pFile->Read(&bh, 8))
	{
		return false;
	}

	if (bh.dwName != BEDW('LECF'))
	{
		ErrorMessage("Invalid SCUMM disk found.");
		return false;
	}

	if (!pFile->Read(&bh, 8))
	{
		return false;
	}

	if (bh.dwName != BEDW('LOFF'))
	{
		ErrorMessage("Invalid SCUMM disk found.");
		return false;
	}

	if (!ParseLOFF(pFile))
	{
		return false;
	}

	return true;
}

void ScummDisk::Destroy()
{
	if (m_pRoomOffsets)
	{
		delete[] m_pRoomOffsets;
		m_pRoomOffsets = NULL;
	}
}

bool ScummDisk::SeekRoom(File* pFile, BYTE bRoom)
{
	if (!pFile->SeekSet(m_pRoomOffsets[bRoom]))
	{
		return false;
	}

	return true;
}

bool ScummDisk::ParseLOFF(File* pFile)
{
	if (!pFile->Read(&m_numRooms, 1))
	{
		return false;
	}

	m_pRoomOffsets = new DWORD[256];
	if (!m_pRoomOffsets)
	{
		ErrorMessage("Out of memory.");
		return false;
	}

	for (BYTE b = 0; b < m_numRooms; b++)
	{
		BYTE bRoom;
		if (!pFile->Read(&bRoom, 1))
		{
			return false;
		}
		if (!pFile->Read(&m_pRoomOffsets[bRoom], 4))
		{
			return false;
		}
	}

	return true;
}

bool ScummDirectory::Parse(bool fScummV8, BYTE* pData)
{
	if (fScummV8)
	{
		m_dwNumItems = *(DWORD*)pData;
		pData += 4;
	}
	else
	{
		m_dwNumItems = *(WORD*)pData;
		pData += 2;
	}

	m_pRooms = new BYTE[m_dwNumItems];
	if (!m_pRooms)
	{
		ErrorMessage("Out of memory.");
		return false;
	}
	m_pOffsets = new DWORD[m_dwNumItems];
	if (!m_pOffsets)
	{
		ErrorMessage("Out of memory.");
		return false;
	}

	for (DWORD dw = 0; dw < m_dwNumItems; dw++)
	{
		m_pRooms[dw] = *pData++;
	}

	for (DWORD dw = 0; dw < m_dwNumItems; dw++)
	{
		m_pOffsets[dw] = *(DWORD*)pData;
		pData += 4;
	}

	return true;
}

void ScummDirectory::Destroy()
{
	if (m_pOffsets)
	{
		delete[] m_pOffsets;
		m_pOffsets = NULL;
	}
	if (m_pRooms)
	{
		delete[] m_pRooms;
		m_pRooms = NULL;
	}
	m_dwNumItems = 0;
}

bool ScummResourceManager::Initialize()
{
	if (!DetectDisks())
	{
		return false;
	}

	bool fDataDisksPresent = false;
	for (int i = 0; i < 10; i++)
	{
		if (m_pDisks[i])
		{
			fDataDisksPresent = true;
		}
	}
	if (!fDataDisksPresent)
	{
		ErrorMessage("No SCUMM disks found.");
		return false;
	}

	return true;
}

void ScummResourceManager::Destroy()
{
	m_dirCostumes.Destroy();
	m_dirRooms.Destroy();
	for (int i = 0; i < 10; i++)
	{
		if (m_pScummDisks[i])
		{
			m_pScummDisks[i]->Destroy();
			delete m_pScummDisks[i];
			m_pScummDisks[i] = NULL;
		}
		if (m_pDisks[i])
		{
			m_pDisks[i]->Close();
			delete m_pDisks[i];
			m_pDisks[i] = NULL;
		}
	}
}

LIRETURNCODE ScummResourceManager::LoadCostume(BYTE*& pMem, DWORD dwCostume)
{
	BYTE bRoom = m_dirCostumes.GetItemRoom(dwCostume);
	if (bRoom == -1)
	{
		return LI_DISK_NOT_PRESENT;
	}

	BYTE bDisk = m_dirRooms.GetItemRoom(bRoom);
	if (!bDisk)
	{
		return LI_DISK_NOT_PRESENT;
	}

	if (bDisk > 9)
	{
		ErrorMessage("Invalid disk.");
		return LI_FATAL_ERROR;
	}

	if (!m_pDisks[bDisk])
	{
		return LI_DISK_NOT_PRESENT;
	}

	if (!m_pScummDisks[bDisk]->SeekRoom(m_pDisks[bDisk], bRoom))
	{
		return LI_FATAL_ERROR;
	}

	if (!m_pDisks[bDisk]->SeekCurrent(m_dirCostumes.GetItemOffset(dwCostume)))
	{
		return LI_FATAL_ERROR;
	}

	BLOCKHEADER bh;
	if (!m_pDisks[bDisk]->Read(&bh, 8))
	{
		return LI_FATAL_ERROR;
	}
	if (bh.dwName != BEDW('AKOS'))
	{
		ErrorMessage("AKOS not found.");
		return LI_FATAL_ERROR;
	}

	if (!m_pDisks[bDisk]->SeekCurrent(-8))
	{
		return LI_FATAL_ERROR;
	}

	pMem = (BYTE*)malloc(BEDW(bh.dwSize));
	if (!pMem)
	{
		ErrorMessage("Out of memory.");
		return LI_FATAL_ERROR;
	}

	if (!m_pDisks[bDisk]->Read(pMem, BEDW(bh.dwSize)))
	{
		ErrorMessage("Out of memory.");
		free(pMem);
		return LI_FATAL_ERROR;
	}

	return LI_SUCCESS;
}

bool ScummResourceManager::DetectDisks()
{
	for (int i = 0; i < 10; i++)
	{
		TCHAR  cPath[MAX_PATH];
		LPTSTR lpFilePart;
		TCHAR  cExtension[5];

		wsprintf(cExtension, ".LA%.1d", i);
		if (SearchPath(m_lpScummPath, m_lpScummGameName, cExtension, MAX_PATH, cPath, &lpFilePart))
		{
			if (i == 0)
			{
				m_pDisks[0] = new File;
				if (!m_pDisks[0])
				{
					ErrorMessage("Out of memory.");
					return false;
				}
				if (!m_pDisks[0]->Open(cPath))
				{
					return false;
				}
				if (!ParseIndexDisk())
				{
					return false;
				}
			}
			else
			{
				m_pDisks[i] = new File;
				if (!m_pDisks[i])
				{
					ErrorMessage("Out of memory.");
					return false;
				}
				if (!m_pDisks[i]->Open(cPath))
				{
					return false;
				}

				m_pScummDisks[i] = new ScummDisk;
				if (!m_pScummDisks[i])
				{
					ErrorMessage("Out of memory.");
					return false;
				}
				if (!m_pScummDisks[i]->Initialize(m_pDisks[i]))
				{
					return false;
				}
			}
		}
	}

	return true;
}

bool ScummResourceManager::ParseIndexDisk()
{
	DWORD dwCursor = 0;
	do
	{
		BLOCKHEADER bh;
		if (!m_pDisks[0]->Read(&bh, 8))
		{
			return false;
		}

		BYTE* pDROOBlock;
		BYTE* pDCOSBlock;
		DWORD dwOriginalCursor;
		switch (bh.dwName)
		{
		case BEDW('RNAM'):
			break;
		case BEDW('MAXS'):
			break;
		case BEDW('DRSC'):
			break;
		case BEDW('DROO'):
			dwOriginalCursor = m_pDisks[0]->GetCursor();
			pDROOBlock = (BYTE*)malloc(BEDW(bh.dwSize) - 8);
			if (!pDROOBlock)
			{
				ErrorMessage("Out of memory.");
				return false;
			}
			if (!m_pDisks[0]->Read(pDROOBlock, BEDW(bh.dwSize) - 8))
			{
				free(pDROOBlock);
				return false;
			}
			if (!m_dirRooms.Parse(true, pDROOBlock))
			{
				free(pDROOBlock);
				return false;
			}
			free(pDROOBlock);
			if (!m_pDisks[0]->SeekSet(dwOriginalCursor))
			{
				return false;
			}
			break;
		case BEDW('DSCR'):
			break;
		case BEDW('DSOU'):
			break;
		case BEDW('DCOS'):
			dwOriginalCursor = m_pDisks[0]->GetCursor();
			pDCOSBlock = (BYTE*)malloc(BEDW(bh.dwSize) - 8);
			if (!pDCOSBlock)
			{
				ErrorMessage("Out of memory.");
				return false;
			}
			if (!m_pDisks[0]->Read(pDCOSBlock, BEDW(bh.dwSize) - 8))
			{
				free(pDCOSBlock);
				return false;
			}
			if (!m_dirCostumes.Parse(true, pDCOSBlock))
			{
				free(pDCOSBlock);
				return false;
			}
			free(pDCOSBlock);
			if (!m_pDisks[0]->SeekSet(dwOriginalCursor))
			{
				return false;
			}
			break;
		case BEDW('DCHR'):
			break;
		case BEDW('DOBJ'):
			break;
		case BEDW('AARY'):
			break;
		case BEDW('ANAM'):
			break;
		default:
			char c[256];
			wsprintf(c, "Unknown block: %c%c%c%c", bh.dwName, bh.dwName >> 8, bh.dwName >> 16, bh.dwName >> 24);
			ErrorMessage(c);
			return false;
		}

		if (!m_pDisks[0]->SeekCurrent(BEDW(bh.dwSize) - 8))
		{
			return false;
		}
		dwCursor += BEDW(bh.dwSize);
	} while (dwCursor < m_pDisks[0]->GetSize());

	return true;
}